#include"rewrite_cache.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>

#define DEBUG(x)
void
hex (unsigned char *str, int len)
{
  int i;
  for (i = 0; i < len; i++)
    printf ("%02x", str[i]);
  printf ("\n");
}

unsigned int
fmtsize (const char *fmt)
{
  const char *p = fmt;
  unsigned int size = 0;
  for (p = fmt; *p; p++)
    switch (*p)
      {
      case 'i':
        size += sizeof (int);
        break;
      case '?':
      case 'p':
        size += sizeof (void *);
        if (*(p + 1) == '[')
          {
            size += sizeof (int);
            while (*(++p) != ']');
          }
        break;
      default:
        break;
      }
  return size;
}

void *
vcreateNode (const char *const fmt, va_list ap)
{
  void *payload = NULL;
  void *payload_p = NULL;
  const char *p = fmt;

  /*calculate the number of items in the fmt */
  unsigned int size = fmtsize (fmt);
  /*alloc the mem for it */
  payload = payload_p = malloc (size);
  assert (payload);

  /*load it up */
  while (*p)
    {
      switch (*p)
        {
        case '?':

          /*Pointer */
          assert (payload_p <= payload + size);
          void *buffer = va_arg (ap, void *);
          *(void **) payload_p = buffer;
          payload_p += sizeof (void *);
          int len = 0;
          if (*(p + 1) == '[')
            {
              p += 2;
              len = isdigit (*p) ? atoi (p) : va_arg (ap, int);
              while (*(++p) != ']');
              *(int *) payload_p = len;
              payload_p += sizeof (int);
            }
          break;
        case 'p':
          {
            /*Pointer */
            assert (payload_p <= payload + size);
            void *buffer = va_arg (ap, void *);
            int len = 0;
            if (*(p + 1) == '[')
              {
                p += 2;
                len = isdigit (*p) ? atoi (p) : va_arg (ap, int);
                while (*(++p) != ']');
                if (buffer)
                  {
                    *(void **) payload_p = malloc (len);
                    memcpy (*(void **) payload_p, buffer, len);
                  }
                payload_p += sizeof (void *);
                *(int *) payload_p = len;
                payload_p += sizeof (int);
              }
            else
              {
                *(void **) payload_p = buffer;
                payload_p += sizeof (void *);
              }
          }
          break;
        case 'i':
          /*Number */
          assert (payload_p <= payload + size);
          *(int *) payload_p = va_arg (ap, int);
          payload_p += sizeof (int);
          break;
        default:
          abort ();                //Not implemented
          break;
        }
      p++;
    }

  struct node *node = malloc (sizeof (struct node));
  node->next = NULL;
  strncpy (node->fmt, fmt, 256);
  node->payload = payload;
  return node;
}


int
eqNode (struct node *a, struct node *b)
{
  assert (a && b);
  if (a == b)
    return 1;

  DEBUG (printf ("Comparing:\n"));
  DEBUG (printNode (a));
  DEBUG (printNode (b));

  assert (fmtsize (a->fmt) == fmtsize (b->fmt));

  char *fmta = a->fmt;
  char *fmtb = b->fmt;
  void *payloada = a->payload;
  void *payloadb = b->payload;

  while (*fmta && *fmtb)
    {
      switch (*fmta)
        {
        case 'i':
          if (*fmtb == '?')
            **(int **) payloadb = *(int *) payloada;
          else if (*(int *) payloada != *(int *) payloadb)
            return 0;
          payloada += sizeof (int);
          payloadb += sizeof (int);
          break;
        case 'p':
          {
            char q = *fmtb;
            void **buffera = (void **) payloada;
            void **bufferb = (void **) payloadb;
            payloada += sizeof (void *);
            payloadb += sizeof (void *);
            assert (*(++fmta) == '[' && *(++fmtb) == '[');
            unsigned int lena =
              isdigit (*fmta) ? atoi (fmta) : *(int *) payloada;
            unsigned int lenb;

            if (*(fmtb + 1) == '?')
              lenb = **(int **) payloadb = lena;
            else
              {
                lenb =
                  isdigit (*fmtb + 1) ? atoi (fmtb + 1) : *(int *) payloadb;
                if (lena != lenb)
                  return 0;
              }

            if (q == '?')
              memcpy (*bufferb, *buffera, lena);
            else if (bufferb && memcmp (*buffera, *bufferb, lena) != 0)
              return 0;

            if (!isdigit (*fmta))
              payloada += sizeof (int);
            if (!isdigit (*fmtb))
              payloadb += sizeof (int);

            while (*(++fmta) != ']');
            while (*(++fmtb) != ']');
          }
          break;
        }
      fmta++;
      fmtb++;
    }

  return 1;
}


void
putNode (struct node **table, const char *const fmt, ...)
{
  struct node *node = NULL;
  struct node **p = table;
  struct node *n;
  assert (table);
  //Iter to the end...
  while (*p != NULL)
    p = &((*p)->next);

  //Create the Node
  va_list ap;
  va_start (ap, fmt);
  node = vcreateNode (fmt, ap);
  va_end (ap);
  //Link it
  DEBUG (printf ("Adding Node:"));
  DEBUG (printNode (node));
  *p = node;
};

void
printNode (struct node *node)
{
  void *payload = node->payload;
  const char *p = node->fmt;
  void *payload_p = payload;
  /*calculate the number of items in the fmt */
  unsigned int size = fmtsize (node->fmt);
  printf ("Node '%s' (%p):\n", node->fmt, node);
  while (*p)
    {
      printf ("\t");
      switch (*p)
        {
        case '?':
          {
            int len = 0;
            printf ("?");
            payload_p += sizeof (void *);
            if (*(p + 1) == '[')
              {
                p += 2;
                if (*p == '?')
                  printf ("[?]:");
                else
                  {
                    len = isdigit (*p) ? atoi (p) : *(int *) payload_p;
                    payload_p += sizeof (int);
                    printf ("[%d]:", len);
                    for (; len > 0; len--)
                      printf ("?");
                  }
                while (*(++p) != ']');
              }
            printf ("\n");
          }
          break;
        case 'p':
          {
            printf ("%c", *p);
            /*Pointer */
            assert (payload_p <= payload + size);
            void *buffer = *(void **) payload_p;
            payload_p += sizeof (void *);
            int len = 0;
            if (*(p + 1) == '[')
              {
                p += 2;
                if (*p == '?')
                  printf ("[?]:");
                else
                  {
                    len = isdigit (*p) ? atoi (p) : *(int *) payload_p;
                    printf ("[%d]:", len);
                  }
                payload_p += sizeof (int);
                while (*(++p) != ']');
              }
            if (buffer)
              {
                hex (buffer, len);
              }
            else
              printf ("\n");
          }
          break;
        case 'i':
          /*Number */
          assert (payload_p <= payload + size);
          printf ("i:%d", *(int *) payload_p);
          payload_p += sizeof (int);
          break;
        default:
          abort ();                //Not implemented
          break;
        }
      p++;
    }
  printf ("\nFMTLen: %d", size);
  printf ("\n");
}


int
findNode (struct node **table, const char *const fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  struct node *node = vcreateNode (fmt, ap);
  va_end (ap);
  assert (table != NULL);
  struct node **p = table;
  DEBUG (printf ("SEARCHING Node:"));
  while (*p != NULL)
    {
      if (eqNode (*p, node))
        {
          DEBUG (printf ("FOUND!!\n"));
          break;
        };
      p = &((*p)->next);
    }
  return (*p != NULL);
}

static int fresh = 0x01;
void
getFresh (char *dst, unsigned int len)
{
  memset (dst, fresh++, len);
}
