#include "rijndael.h"
#include "rewrite_cache.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

//Table for symbolic rijndael_set_key function ctx(>p[i], p[i], i)
static struct node *rijndael_set_key_tbl = NULL;

void
rijndael_set_key (rijndael_ctx * ctx, u_char * key, int bits, int do_encrypt)
{
  //preconditions
  assert (bits == 128 || bits == 192 || bits == 256);
  assert (do_encrypt == 1 || do_encrypt == 0);
  //search the cache, if fount it willl automagically fill the ctx
  if (!findNode
      (&rijndael_set_key_tbl, "?[i]p[i]i", ctx, sizeof (*ctx), key, bits / 8,
       do_encrypt))
    {
      //if not  found get one fresh
      getFresh ((void *) ctx, sizeof (*ctx));
      //save it in the cache
      putNode (&rijndael_set_key_tbl, "p[i]p[i]i", ctx, sizeof (*ctx), key,
               bits / 8, do_encrypt);
    }
}

//Table for symbolic rijndael_encrypt function enc(p[i], p[16], >p[16])
static struct node *rijndael_encrypt_tbl = NULL;

void
rijndael_encrypt (rijndael_ctx * ctx, u_char * src, u_char * dst)
{
  //search the cache, if fount it will automagically fill the dst
  if (!findNode
      (&rijndael_encrypt_tbl, "p[i]p[16]?[16]", ctx, sizeof (*ctx), src, dst))
    {
      //if not  found get one fresh
      getFresh (dst, 16);
      //save it in the cache
      putNode (&rijndael_encrypt_tbl, "p[i]p[16]p[16]", ctx, sizeof (*ctx),
               src, dst);
    }
}

//Table for symbolic rijndael_encrypt function dec(p[i], p[16], >p[16])
static struct node *rijndael_decrypt_tbl = NULL;

void
rijndael_decrypt (rijndael_ctx * ctx, u_char * src, u_char * dst)
{
  //search the cache, if fount it will automagically fill the dst
  if (findNode (&rijndael_decrypt_tbl, "p[i]p[16]?[16]", ctx, sizeof (*ctx), src, dst))
    return;

  { //Get the context and the clear text that produces the encrypted text in src
    rijndael_ctx enc_ctx;
    if (!findNode (&rijndael_encrypt_tbl, "?[i]?[16]p[16]", &enc_ctx, sizeof (*ctx), dst, src))
      goto fresh;                //If not found get a fresh symbolic value

    {
      /** 
       * Because we can encrypt with ctx(K,n,1) or ctx(K,n,0) but decrypt 
       * only with ctx(K,n,0) search rijndael_set_key_tbl table to get the  
       * params, key and key_len used to generate that context
       */
      u_char enc_key[32];
      int enc_bits_8;
      int enc_do_encrypt;
      if (!findNode
          (&rijndael_set_key_tbl, "p[i]?[?]?", &enc_ctx, sizeof (enc_ctx),
           enc_key, &enc_bits_8, &enc_do_encrypt))
        goto fresh; //If not found get a fresh symbolic value
      {
        //Now try to get the params used to generate the context passed
        u_char key[32];
        int bits_8;
        int do_encrypt;
        if (!findNode
            (&rijndael_set_key_tbl, "p[i]?[?]?", ctx, sizeof (*ctx), &key,
             &bits_8, &do_encrypt))
          goto fresh;//If not found get a fresh symbolic value

        //compare both contexts to by compatible
        if (!
            (memcmp (enc_key, key, 16) == 0 && bits_8 == enc_bits_8
             && enc_do_encrypt == 0))
          goto fresh;                //If not found get a fresh symbolic value
      }
    }
  }
  return;
fresh:
  // Something in the rewriting went inconclusive .. so we make a new fresh symbol out of this call
  getFresh (dst, 16);
  putNode (&rijndael_decrypt_tbl, "p[i]p[16]p[16]", ctx, sizeof (*ctx), src,
           dst);
}
